from typing import Optional

from ape.api import PluginConfig
from pydantic_settings import SettingsConfigDict

from .types import SafeCacheData


class SafeConfig(PluginConfig):
    default_safe: Optional[str] = None
    """Alias of the default safe."""

    require: dict[str, SafeCacheData] = {}
    """Safes that are required to exist for the project. Useful for cloud-based usage."""

    model_config = SettingsConfigDict(env_prefix="APE_SAFE_")



import pytest

from ape_safe.exceptions import SafeClientException


def test_manage_delegates(safe, delegate, OWNERS):
    owner = OWNERS[0]
    assert owner.address not in safe.client.get_delegates()

    safe.client.add_delegate(delegate.address, "pepito", owner)
    assert delegate.address in safe.client.get_delegates()[owner.address]
    assert delegate.address in safe.all_delegates()
    # NOTE: Only in MockSafeClient
    assert safe.client.delegator_for_delegate(delegate.address) == owner.address

    safe.client.remove_delegate(delegate.address, owner)
    assert owner.address not in safe.client.get_delegates()

    with pytest.raises(SafeClientException):
        # Only signers can create a delegate
        safe.client.add_delegate(owner, "root privledges", delegate)


def test_delegate_can_propose_safe_tx(safe, delegate, OWNERS):
    owner = OWNERS[0]

    safe_tx = safe.create_safe_tx(
        safe.contract.addOwnerWithThreshold.as_transaction(delegate, safe.confirmations_required)
    )

    with pytest.raises(SafeClientException):
        # Not a delegate or signer
        safe.propose_safe_tx(safe_tx, submitter=delegate)

    safe.client.add_delegate(delegate.address, "pepito", owner)
    safe.propose_safe_tx(safe_tx, submitter=delegate)

    assert len(safe.get_api_confirmations(safe_tx)) == 0
    assert list(
        (safe_tx.signable_message, confs) for safe_tx, confs in safe.pending_transactions()
    ) == [(safe_tx.signable_message, [])]



from ape.types.address import AddressType
from pydantic import BaseModel


class SafeCacheData(BaseModel):
    """Model for cached Safe data under `~/.ape/safe/*.json`"""

    address: AddressType
    """Address of the Safe"""

    deployed_chain_ids: list[int] = []
    """Networks (by chain ID) this Safe is deployed on"""



import pytest
from ape.api import AccountAPI
from ape.types import AddressType


def test_init(safe, OWNERS, THRESHOLD, safe_contract):
    assert safe.contract == safe_contract
    assert safe.confirmations_required == THRESHOLD
    assert safe.signers == list(o.address for o in OWNERS)
    assert safe.next_nonce == 0


def test_swap_owner(safe, accounts, OWNERS, exec_transaction):
    old_owner = safe.signers[0]
    new_owner = accounts[len(OWNERS)]  # replace owner 1 with account N + 1
    assert new_owner.address not in safe.signers
    # NOTE: Since the signers are processed in order, we replace the last account

    prev_owner = safe.compute_prev_signer(old_owner)

    receipt = exec_transaction(
        safe.contract.swapOwner,
        prev_owner,
        old_owner,
        new_owner,
    )

    assert receipt.events == [
        safe.contract.RemovedOwner(owner=old_owner),
        safe.contract.AddedOwner(owner=new_owner),
        safe.contract.ExecutionSuccess(),
    ]

    assert old_owner not in safe.signers
    assert new_owner.address in safe.signers


def test_add_owner(safe, accounts, OWNERS, exec_transaction):
    new_owner = accounts[len(OWNERS)]  # replace owner 1 with account N + 1
    assert new_owner.address not in safe.signers

    receipt = exec_transaction(
        safe.contract.addOwnerWithThreshold,
        new_owner,
        safe.confirmations_required,
    )

    assert receipt.events == [
        safe.contract.AddedOwner(owner=new_owner),
        safe.contract.ExecutionSuccess(),
    ]

    assert new_owner.address in safe.signers


@pytest.mark.parametrize("mode", ["impersonate", "api", "sign"])
def test_remove_owner(safe, OWNERS, exec_transaction):
    if len(OWNERS) == 1:
        pytest.skip("Can't remove the only owner")

    old_owner = safe.signers[0]
    new_threshold = max(len(OWNERS) - 1, safe.confirmations_required - 1)
    threshold_changed = new_threshold != safe.confirmations_required

    prev_owner = safe.compute_prev_signer(old_owner)

    receipt = exec_transaction(
        safe.contract.removeOwner,
        prev_owner,
        old_owner,
        # Can't set the threshold to zero or more than the number of owners after removal
        new_threshold,
    )

    expected_events = [
        safe.contract.RemovedOwner(owner=old_owner),
        safe.contract.ExecutionSuccess(),
    ]
    if threshold_changed:
        expected_events.insert(1, safe.contract.ChangedThreshold(threshold=new_threshold))
    assert receipt.events == expected_events

    assert old_owner not in safe.signers


def test_account_type(safe):
    actual = type(safe)
    assert issubclass(actual, AccountAPI)


def test_safe_account_convert(safe):
    """
    We had a bug where converting safe accounts to AddressType
    would fail.
    """
    convert = safe.conversion_manager.convert
    actual = convert(safe, AddressType)
    assert actual == safe.address


def test_get_client(safe):
    """
    Test getting a client for a specific chain ID.
    """
    # Create a test deployed_chain_ids list
    safe_data = safe.account_data
    safe_data.deployed_chain_ids = [1, 10, 100]
    safe.account_file_path.write_text(safe_data.model_dump_json())

    # Should work for a specified chain ID
    client = safe.get_client(chain_id=1)
    assert "/eth/" in client.base_url



from collections.abc import Iterator
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Optional, Union, cast

from ape.utils import ZERO_ADDRESS, ManagerAccessMixin
from eth_utils import keccak, to_hex
from hexbytes import HexBytes

from ape_safe.client.base import BaseSafeClient
from ape_safe.client.types import (
    SafeApiTxData,
    SafeDetails,
    SafeTx,
    SafeTxConfirmation,
    SafeTxID,
    SignatureType,
    UnexecutedTxData,
)
from ape_safe.exceptions import SafeClientException
from ape_safe.utils import get_safe_tx_hash

if TYPE_CHECKING:
    from ape.api import AccountAPI
    from ape.contracts import ContractInstance
    from ape.types import AddressType, MessageSignature


class MockSafeClient(BaseSafeClient, ManagerAccessMixin):
    def __init__(self, contract: "ContractInstance"):
        self.contract = contract
        self.transactions: dict[SafeTxID, SafeApiTxData] = {}
        self.transactions_by_nonce: dict[int, list[SafeTxID]] = {}
        self.delegates: dict["AddressType", list["AddressType"]] = {}

    @property
    def safe_details(self) -> SafeDetails:
        slot = keccak(text="fallback_manager.handler.address")
        value = self.provider.get_storage(self.contract.address, slot)
        fallback_address = self.network_manager.ecosystem.decode_address(value[-20:])

        return SafeDetails(
            address=self.contract.address,
            nonce=self.get_next_nonce(),
            threshold=self.contract.getThreshold(),
            owners=self.contract.getOwners(),
            masterCopy=self.contract.masterCopy(),
            modules=self.modules,
            # TODO: Add fallback handler getter
            fallbackHandler=fallback_address,
            guard=self.guard,
            version=self.contract.VERSION(),
        )

    @property
    def guard(self) -> "AddressType":
        return (
            self.contract.getGuard() if "getGuard" in self.contract._view_methods_ else ZERO_ADDRESS
        )

    @property
    def modules(self) -> list["AddressType"]:
        return self.contract.getModules() if "getModules" in self.contract._view_methods_ else []

    def get_next_nonce(self) -> int:
        return self.contract._view_methods_["nonce"]()

    def _all_transactions(
        self,
    ) -> Iterator[SafeApiTxData]:
        for nonce in sorted(self.transactions_by_nonce.keys(), reverse=True):
            for tx in map(self.transactions.get, self.transactions_by_nonce[nonce]):
                if tx:
                    yield tx

    def get_safe_tx(self, safe_tx_hash: SafeTxID) -> SafeApiTxData:
        tx_hash = cast(SafeTxID, to_hex(HexBytes(safe_tx_hash)))

        if safe_tx := self.transactions.get(tx_hash):
            return safe_tx

        for safe_tx in self._all_transactions():
            if safe_tx.safe_tx_hash == safe_tx_hash:
                return safe_tx

        raise SafeClientException(f"Unable to find SafeTx '{safe_tx_hash}'.")

    def get_confirmations(self, safe_tx_hash: SafeTxID) -> Iterator[SafeTxConfirmation]:
        tx_hash = cast(SafeTxID, to_hex(HexBytes(safe_tx_hash)))
        if safe_tx_data := self.transactions.get(tx_hash):
            yield from safe_tx_data.confirmations

    def post_transaction(
        self, safe_tx: SafeTx, signatures: dict["AddressType", "MessageSignature"], **kwargs
    ):
        owners = self.safe_details.owners

        safe_tx_data = UnexecutedTxData.from_safe_tx(safe_tx, self.safe_details.threshold)
        safe_tx_data.confirmations.extend(
            SafeTxConfirmation(
                owner=signer,
                submissionDate=datetime.now(timezone.utc),
                signature=sig.encode_rsv(),
                signatureType=SignatureType.EOA,
            )
            for signer, sig in signatures.items()
            if signer in owners
        )

        # Ensure that if this is a zero-conf SafeTx, that at least one signature is from a delegate
        # NOTE: More strict than Safe API check that silently ignores if no signatures are valid or
        #       from delegates, but should help us to get correct logic for mock testing purposes
        if len(safe_tx_data.confirmations) == 0 and not any(
            self.delegator_for_delegate(delegate) in owners
            for delegate in filter(lambda signer: signer not in owners, signatures)
        ):
            # NOTE: mimic real exception for mock testing purposes
            raise SafeClientException(
                "At least one signature must be from a valid owner of the safe"
            )

        tx_id = cast(SafeTxID, to_hex(HexBytes(safe_tx_data.safe_tx_hash)))
        self.transactions[tx_id] = safe_tx_data
        if safe_tx_data.nonce in self.transactions_by_nonce:
            self.transactions_by_nonce[safe_tx_data.nonce].append(tx_id)
        else:
            self.transactions_by_nonce[safe_tx_data.nonce] = [tx_id]

    def post_signatures(
        self,
        safe_tx_or_hash: Union[SafeTx, SafeTxID],
        signatures: dict["AddressType", "MessageSignature"],
    ):
        for signer, signature in signatures.items():
            safe_tx_id = (
                safe_tx_or_hash
                if isinstance(safe_tx_or_hash, (str, bytes, int))
                else get_safe_tx_hash(safe_tx_or_hash)
            )
            tx_id = cast(SafeTxID, to_hex(HexBytes(safe_tx_id)))
            self.transactions[tx_id].confirmations.append(
                SafeTxConfirmation(
                    owner=signer,
                    submissionDate=datetime.now(timezone.utc),
                    signature=signature.encode_rsv(),
                    signatureType=SignatureType.EOA,
                )
            )

    def estimate_gas_cost(
        self, receiver: "AddressType", value: int, data: bytes, operation: int = 0
    ) -> Optional[int]:
        return None  # Estimate gas normally

    def get_delegates(self) -> dict["AddressType", list["AddressType"]]:
        return self.delegates

    def delegator_for_delegate(self, delegate: "AddressType") -> Optional["AddressType"]:
        for delegator, delegates in self.delegates.items():
            if delegate in delegates:
                return delegator

        return None

    def add_delegate(self, delegate: "AddressType", label: str, delegator: "AccountAPI"):
        if delegator.address not in self.safe_details.owners:
            raise SafeClientException(f"'{delegator}' not a valid owner.")

        if delegator.address in self.delegates:
            self.delegates[delegator.address].append(delegate)

        else:
            self.delegates[delegator.address] = [delegate]

    def remove_delegate(self, delegate: "AddressType", delegator: "AccountAPI"):
        if delegator.address not in self.safe_details.owners:
            raise SafeClientException(f"'{delegator.address}' not a valid owner.")

        elif delegator.address not in self.delegates:
            raise SafeClientException(
                f"'{delegate}' not a valid delegate for '{delegator.address}'."
            )

        self.delegates[delegator.address].remove(delegate)

        if len(self.delegates[delegator.address]) == 0:
            del self.delegates[delegator.address]



import click
from ape.cli import ConnectedProviderCommand, account_option
from ape.types import AddressType
from eth_typing import ChecksumAddress

from ape_safe._cli.click_ext import safe_argument, safe_cli_ctx, safe_option


@click.group()
def delegates():
    """
    View and configure delegates
    """


@delegates.command("list", cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_argument
def _list(cli_ctx, safe):
    """
    Show delegates for signers in a Safe
    """
    if delegates := safe.client.get_delegates():
        cli_ctx.logger.success(f"Found delegates for {safe.address} ({safe.alias})")
        for delegator in delegates:
            click.echo(f"\nSigner {delegator}:")
            click.echo("- " + "\n- ".join(delegates[delegator]))

    else:
        cli_ctx.logger.info(f"No delegates for {safe.address} ({safe.alias})")


@delegates.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@click.argument("delegate", type=ChecksumAddress)
@click.argument("label")
@account_option()
def add(cli_ctx, safe, delegate, label, account):
    """
    Add a delegate for a signer in a Safe
    """
    delegate = cli_ctx.conversion_manager.convert(delegate, AddressType)
    safe.client.add_delegate(delegate, label, account)
    cli_ctx.logger.success(
        f"Added delegate {delegate} ({label}) for {account.address} "
        f"in {safe.address} ({safe.alias})"
    )


@delegates.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@safe_option
@click.argument("delegate", type=ChecksumAddress)
@account_option()
def remove(cli_ctx, safe, delegate, account):
    """
    Remove a delegate for a specific signer in a Safe
    """
    delegate = cli_ctx.conversion_manager.convert(delegate, AddressType)
    safe.client.remove_delegate(delegate, account)
    cli_ctx.logger.success(
        f"Removed delegate {delegate} for {account.address} in {safe.address} ({safe.alias})"
    )



from enum import Enum
from functools import cache
from importlib import resources
from typing import TYPE_CHECKING, Union

import requests
from ape.managers.project import ProjectManager
from packaging.version import Version
from pydantic import BaseModel

if TYPE_CHECKING:
    from ape.contracts import ContractContainer, ContractInstance


with resources.as_file(resources.files(__package__).joinpath("manifests")) as manifest_folder:
    MANIFESTS_BY_VERSION = {
        Version(m.stem.lstrip("safe-v")): m for m in manifest_folder.glob("safe-v*.json")
    }


@cache
def get_manifest(version: Version) -> ProjectManager:
    if not (manifest_path := MANIFESTS_BY_VERSION.get(version)):
        raise KeyError(f"Unknown version 'v{version}'.")

    return ProjectManager.from_manifest(manifest_path)


class PackageType(str, Enum):
    SINGLETON = "SafeSingleton"
    PROXY = "SafeProxy"
    PROXY_FACTORY = "SafeProxyFactory"
    MULTISEND = "MultiSend"

    def __call__(self, version: Union[Version, str]) -> "ContractContainer":
        if not isinstance(version, Version):
            version = Version(version.lstrip("v"))

        package = get_manifest(version)

        if self is PackageType.MULTISEND:
            if version == Version("1.1.1"):
                return package.MultiSend

            # NOTE: Always use `MultiSendCallOnly` to prevent against delegatecall issues
            return package.MultiSendCallOnly

        elif self is PackageType.PROXY_FACTORY:
            if version == Version("1.1.1"):
                return package.ProxyFactory

            elif version <= Version("1.3.0"):
                return package.GnosisSafeProxyFactory

            else:
                return package.SafeProxyFactory

        elif version > Version("1.3.0"):
            SafeSingleton = package.Safe

        else:
            SafeSingleton = package.GnosisSafe

        if self is PackageType.SINGLETON:
            return SafeSingleton

        # NOTE: SafeProxy has `masterCopy() -> address` in it, vs. `SafeSingleton`
        SafeSingleton.contract_type.abi.append(package.IProxy.contract_type.abi[0])
        return SafeSingleton


class DeploymentType(str, Enum):
    CANONICAL = "canonical"
    EIP155 = "eip155"
    ZKSYNC = "zksync"


class DeploymentInfo(BaseModel):
    address: str
    codeHash: str


class DeploymentAsset(BaseModel):
    released: bool
    contractName: str
    version: str
    deployments: dict[DeploymentType, DeploymentInfo]
    # ChainID => DeploymentType
    networkAddresses: dict[int, Union[DeploymentType, list[DeploymentType]]]


BASE_ASSETS_URL = (
    "https://raw.githubusercontent.com/safe-global/safe-deployments/refs/heads/main/src/assets/"
)


def get_deployment_artifact(
    package_type: PackageType, chain_id: int, version: Union[Version, str]
) -> "ContractInstance":
    if not isinstance(version, Version):
        version = Version(version.lstrip("v"))

    if package_type is PackageType.SINGLETON:
        deployment_filename = "gnosis_safe.json" if version <= Version("1.3.0") else "/safe.json"

    elif package_type is PackageType.PROXY_FACTORY:
        deployment_filename = (
            "proxy_factory.json" if version <= Version("1.3.0") else "/safe_proxy_factory.json"
        )

    elif package_type is PackageType.MULTISEND:
        deployment_filename = "multi_send.json"

    else:
        raise

    # TODO: Cache this to disk?
    response = requests.get(BASE_ASSETS_URL + f"v{version}/{deployment_filename}")
    deployment_asset = DeploymentAsset.model_validate(response.json())

    if not (deployment_type := deployment_asset.networkAddresses.get(chain_id)):
        raise KeyError(f"Chain ID '{chain_id}' does not have a known deployment")

    deployment_info = deployment_asset.deployments[
        deployment_type[0] if isinstance(deployment_type, list) else deployment_type
    ]

    return package_type(version).at(deployment_info.address)  # type: ignore[misc]


def get_singleton(chain_id: int, version: Union[Version, str]) -> "ContractInstance":
    return get_deployment_artifact(PackageType.SINGLETON, chain_id, version)


def get_factory(chain_id: int, version: Union[Version, str]) -> "ContractInstance":
    return get_deployment_artifact(PackageType.PROXY_FACTORY, chain_id, version)


def get_multisend(chain_id: int, version: Union[Version, str]) -> "ContractInstance":
    return get_deployment_artifact(PackageType.MULTISEND, chain_id, version)



import secrets
from functools import cache
from typing import TYPE_CHECKING, Iterable, Optional, Union

from ape.types import AddressType
from ape.utils import ZERO_ADDRESS, ManagerAccessMixin
from packaging.version import Version

from ape_safe.packages import MANIFESTS_BY_VERSION, PackageType, get_factory, get_singleton

if TYPE_CHECKING:
    from ape.api import AccountAPI, BaseAddress
    from ape.contracts import ContractInstance


class SafeFactory(ManagerAccessMixin):
    _singleton: dict[Version, "ContractInstance"] = {}
    _factory: dict[Version, "ContractInstance"] = {}

    @classmethod
    def inject(cls, version: Version, deployer: "AccountAPI"):
        cls._singleton[version] = deployer.deploy(PackageType.SINGLETON(version))
        cls._factory[version] = deployer.deploy(PackageType.PROXY_FACTORY(version))

    @cache
    def get_factory(self, version: Version) -> "ContractInstance":
        if injected_factory := self._factory.get(version):
            return injected_factory

        return get_factory(self.chain_manager.chain_id, version)

    @property
    def contract(self) -> "ContractInstance":
        return self.get_factory(max(MANIFESTS_BY_VERSION))

    @cache
    def get_singleton(self, version: Version) -> "ContractInstance":
        if injected_singleton := self._singleton.get(version):
            return injected_singleton

        return get_singleton(self.chain_manager.chain_id, version)

    def create(
        self,
        owners: Iterable[Union["BaseAddress", "AddressType", str]],
        threshold: int,
        callback_address: Union["BaseAddress", "AddressType", str] = ZERO_ADDRESS,
        callback_calldata: Optional[bytes] = None,
        fallback_handler: Union["BaseAddress", "AddressType", str] = ZERO_ADDRESS,
        payment_token: Union["BaseAddress", "AddressType", str] = ZERO_ADDRESS,
        payment_amount: Union[str, int] = 0,
        payment_receiver: Union["BaseAddress", "AddressType", str] = ZERO_ADDRESS,
        salt: Optional[int] = None,
        version: Union[Version, str, None] = None,
        **txn_kwargs,
    ) -> "ContractInstance":
        if not (owners := [self.conversion_manager.convert(a, AddressType) for a in owners]):
            raise ValueError("Cannot make a Safe with 0 owners.")

        elif not (1 <= threshold <= len(owners)):
            raise ValueError(f"Threshold must be between '1' and '{len(owners)}'")

        if callback_address != ZERO_ADDRESS:
            callback_address = self.conversion_manager.convert(callback_address, AddressType)

        if fallback_handler != ZERO_ADDRESS:
            fallback_handler = self.conversion_manager.convert(fallback_handler, AddressType)

        if (payment_amount := self.conversion_manager.convert(payment_amount, int)) > 0 and (
            payment_token == ZERO_ADDRESS or payment_receiver == ZERO_ADDRESS
        ):
            raise ValueError(
                "If sending payments, must include both `payment_token` and `payment_receiver`"
            )

        else:  # Both are not empty
            payment_token = self.conversion_manager.convert(payment_token, AddressType)
            payment_receiver = self.conversion_manager.convert(payment_receiver, AddressType)

        if not salt:
            salt = secrets.randbits(256)

        if not version:
            version = max(MANIFESTS_BY_VERSION)

        elif not isinstance(version, Version):
            version = Version(version.lstrip("v"))

        Proxy = PackageType.PROXY(version)

        proxy_setup_abi = next(abi for abi in Proxy.abi if getattr(abi, "name", None) == "setup")
        args = self.conversion_manager.convert_method_args(
            proxy_setup_abi,
            [
                owners,
                threshold,
                callback_address,
                callback_calldata or b"",
                fallback_handler,
                payment_token,
                payment_amount,
                payment_receiver,
            ],
        )
        encoded_args = self.provider.network.ecosystem.encode_calldata(proxy_setup_abi, *args)
        method_id = self.provider.network.ecosystem.get_method_selector(proxy_setup_abi)
        initializer_calldata = method_id + encoded_args

        factory_contract = self.get_factory(version)
        singleton = self.get_singleton(version)
        tx = factory_contract.createProxyWithNonce(
            singleton, initializer_calldata, salt, **txn_kwargs
        )

        # NOTE: Use log decoding because it is more available than `.return_value`
        proxy_address = tx.decode_logs(factory_contract.ProxyCreation)[0].proxy
        return Proxy.at(proxy_address)



import ape
import pytest
from ape.utils import create_tempdir
from click.testing import CliRunner

from ape_safe._cli import cli as ape_safe_cli


@pytest.fixture
def safe_container():
    return ape.accounts.containers["safe"]


# NOTE: Every test gets a different data folder
@pytest.fixture(scope="function", autouse=True)
def patch_data_folder(monkeypatch, safe_container):
    with create_tempdir() as data_folder_override:
        monkeypatch.setattr(safe_container, "data_folder", data_folder_override)
        yield


@pytest.fixture
def runner():
    yield CliRunner(mix_stderr=True)


@pytest.fixture
def cli():
    return ape_safe_cli


@pytest.fixture
def safe_account(safe_container, safe):
    safe_container.save_account(safe.alias, safe.address)

    yield safe_container.load_account(safe.alias)

    if safe.alias in safe_container.aliases:
        safe_container.delete_account(safe.alias)



from io import BytesIO
from typing import TYPE_CHECKING, Any

from ape import convert
from ape.types import AddressType, HexBytes
from ape.utils import ManagerAccessMixin, cached_property
from eth_abi.packed import encode_packed

from ape_safe.client.types import OperationType, SafeTxID

from .accounts import SafeAccount, get_signatures
from .exceptions import ApeSafeException, UnsupportedChainError, ValueRequired
from .packages import MANIFESTS_BY_VERSION, PackageType, get_multisend

if TYPE_CHECKING:
    from ape.api import ReceiptAPI, TransactionAPI
    from ape.contracts.base import ContractInstance, ContractTransactionHandler
    from eip712.common import SafeTx
    from packaging.version import Version


class MultiSend(ManagerAccessMixin):
    """
    Create a sequence of calls to execute at once using ``eth_sendTransaction``
    via the MultiSend contract.

    Usage example::

        from ape_safe import multisend
        from ape import accounts

        # load the safe
        safe = accounts.load("my-safe")

        txn = multisend.MultiSend()
        txn.add(contract.myMethod, *call_args)
        txn.add(contract.myMethod, *call_args)
        ...  # Add as many calls as desired to execute
        txn.add(contract.myMethod, *call_args)
        # or, using a builder pattern:
        txn = multisend.MultiSend()
            .add(contract.myMethod, *call_args)
            .add(contract.myMethod, *call_args)
            ...  # Add as many calls as desired to execute
            .add(contract.myMethod, *call_args)

        # Stage the transaction to publish on-chain
        # NOTE: if not enough signers are available, publish to Safe API instead
        receipt = txn(sender=safe,gas=0)
    """

    def __init__(self, version: "Version" = max(MANIFESTS_BY_VERSION)) -> None:
        """
        Initialize a new MultiSend session object. By default, there are no calls to make.
        """
        self.calls: list[dict] = []
        self.version = version

    @classmethod
    def inject(cls, version: "Version" = max(MANIFESTS_BY_VERSION)):
        """
        Create the multisend module contract on-chain, so we can use it.
        Must use a provider that supports ``debug_setCode``.

        Usage example::

            from ape_safe.multisend import MultiSend

            @pytest.fixture(scope="session")
            def multisend():
                MultiSend.inject()
                return MultiSend()
        """
        active_provider = cls.network_manager.active_provider
        assert active_provider, "Must be connected to an active network to deploy"
        MultiSend = PackageType.MULTISEND(version)

        active_provider.set_code(
            "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D",
            MultiSend.contract_type.get_runtime_bytecode(),
        )

    @cached_property
    def contract(self) -> "ContractInstance":
        chain_id = self.network_manager.active_provider.chain_id
        try:
            return get_multisend(chain_id, self.version)

        except KeyError:
            try:
                return PackageType.MULTISEND(self.version).at(
                    "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D"
                )

            except Exception as e:
                raise UnsupportedChainError() from e

    @property
    def handler(self) -> "ContractTransactionHandler":
        return self.contract.multiSend

    def add(
        self,
        call,
        *args,
        value: int = 0,
    ) -> "MultiSend":
        """
        Append a call to the MultiSend session object.

        Raises:
            :class:`InvalidOption`: If one of the kwarg modifiers is not able to be used.

        Args:
            call: :class:`ContractMethodHandler` The method to call.
            *args: The arguments to invoke the method with.
            value: int The amount of ether to forward with the call. Defaults to 0.
            operation: :enum:`OperationType` The type of the operation. Defaults to 0.
        """
        if value < 0:
            raise ValueError("`value=` must be positive.")

        self.calls.append(
            {
                "target": call.contract.address,
                "value": value,
                "callData": call.encode_input(*args),
            }
        )
        return self

    def _validate_safe_tx(self, safe_tx: "SafeTx") -> None:
        required_value = sum(call["value"] for call in self.calls)
        if required_value > safe_tx.value:
            raise ValueRequired(required_value)

    @property
    def encoded_calls(self):
        return [
            encode_packed(
                ["uint8", "address", "uint256", "uint256", "bytes"],
                [
                    # NOTE: Only allow doing CALL because of `MultiSendCallOnly`
                    int(OperationType.CALL),
                    call["target"],
                    call["value"],
                    len(call["callData"]),
                    call["callData"],
                ],
            )
            for call in self.calls
        ]

    def as_safe_tx(self, safe: SafeAccount, **safe_tx_kwargs) -> "SafeTx":
        return safe.safe_tx_def(  # type: ignore[call-arg]
            to=self.contract.address,
            data=self.handler.encode_input(b"".join(self.encoded_calls)),
            operation=OperationType.DELEGATECALL,
            nonce=safe_tx_kwargs.pop("nonce", None) or safe.new_nonce,
            **safe_tx_kwargs,
        )

    def propose(
        self,
        safe: SafeAccount,
        **safe_tx_kwargs,
    ) -> SafeTxID:
        submitter = safe_tx_kwargs.pop("submitter", None)
        sigs_by_signer = safe_tx_kwargs.pop("sigs_by_signer", None)
        safe_tx = self.as_safe_tx(safe, **safe_tx_kwargs)
        self._validate_safe_tx(safe_tx)
        return safe.propose_safe_tx(
            safe_tx,
            submitter=submitter,
            sigs_by_signer=sigs_by_signer,
        )

    def as_transaction(
        self, sender: Any = None, impersonate: bool = False, **txn_kwargs
    ) -> "TransactionAPI":
        """
        Encode the MultiSend transaction as a ``TransactionAPI`` object, but do not execute it.

        Returns:
            :class:`~ape.api.transactions.TransactionAPI`
        """
        if not isinstance(sender, SafeAccount):
            raise ApeSafeException("Must be a SafeAccount to use Multisend")

        safe_tx_kwargs = {}
        for field in sender.safe_tx_def.__annotations__:
            if field in txn_kwargs:
                safe_tx_kwargs[field] = txn_kwargs.pop(field)

        safe_tx = self.as_safe_tx(sender, **safe_tx_kwargs)
        signatures = {} if impersonate else get_signatures(safe_tx, sender.local_signers)
        return sender.create_execute_transaction(safe_tx, signatures=signatures, **txn_kwargs)

    def __call__(self, sender: Any = None, **txn_kwargs) -> "ReceiptAPI":
        """
        Execute the MultiSend transaction. The transaction will broadcast again every time
        the ``Transaction`` object is called.

        Raises:
            :class:`UnsupportedChain`: If there is not an instance of MultiSend deployed
              on the current chain at the expected address.

        Args:
            **txn_kwargs: the kwargs to pass through to the transaction handler.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """
        impersonate = txn_kwargs.pop("impersonate", False)
        return sender.call(
            self.as_transaction(sender=sender, impersonate=impersonate, **txn_kwargs),
            impersonate=impersonate,
            **txn_kwargs,
        )

    def add_from_calldata(self, calldata: bytes):
        """
        Decode all calls from a multisend calldata and add them to this MultiSend.

        Args:
            calldata: Calldata encoding the MultiSend.multiSend call
        """
        _, args = self.contract.decode_input(calldata)
        buffer = BytesIO(args["transactions"])
        while buffer.tell() < len(args["transactions"]):
            operation = int.from_bytes(buffer.read(1), "big")
            target = convert(buffer.read(20), AddressType)
            value = int.from_bytes(buffer.read(32), "big")
            length = int.from_bytes(buffer.read(32), "big")
            data = HexBytes(buffer.read(length))
            self.calls.append(
                {
                    "operation": operation,
                    "target": target,
                    "value": value,
                    "callData": data,
                }
            )



from datetime import datetime

from ape_safe.client import ExecutedTxData


def test_help(runner, cli):
    result = runner.invoke(cli, ["pending", "--help"], catch_exceptions=False)
    assert result.exit_code == 0, result.output


def test_propose(runner, cli, safe_account, receiver, chain):
    nonce_at_start = safe_account.next_nonce
    cmd = (
        "pending",
        "propose",
        "--safe",
        safe_account.alias,
        "--to",
        receiver.address,
        "--value",
        "1",
        "--network",
        chain.provider.network_choice,
    )

    # Sender is required by the API even for initial proposal,
    # so it prompts the user.
    sender_input = f"{safe_account.alias}\n"

    result = runner.invoke(cli, cmd, input=sender_input)
    assert result.exit_code == 0

    # The nonce is the same because we did not execute.
    assert safe_account.next_nonce == nonce_at_start


def test_propose_with_sender(runner, cli, safe_account, receiver, chain):
    # First, fund the safe so the tx does not fail.
    receiver.transfer(safe_account, "1 ETH")

    nonce_at_start = safe_account.next_nonce
    cmd = (
        "pending",
        "propose",
        "--safe",
        safe_account.alias,
        "--to",
        receiver.address,
        "--value",
        "1",
        "--sender",
        receiver.address,
        "--network",
        chain.provider.network_choice,
    )
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert result.exit_code == 0, result.output

    # The nonce is the same because we did not execute.
    assert safe_account.next_nonce == nonce_at_start


def test_propose_with_execute(runner, cli, safe_account, receiver, chain):
    # First, fund the safe so the tx does not fail.
    receiver.transfer(safe_account, "1 ETH")

    nonce_at_start = safe_account.next_nonce
    cmd = (
        "pending",
        "propose",
        "--safe",
        safe_account.alias,
        "--to",
        receiver.address,
        "--value",
        "1",
        "--sender",
        receiver.address,
        "--execute",
        "--network",
        chain.provider.network_choice,
    )
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert safe_account.next_nonce == nonce_at_start + 1


def test_list_no_safes(runner, cli, chain):
    result = runner.invoke(cli, ["pending", "list", "--network", chain.provider.network_choice])
    assert result.exit_code != 0, result.output
    assert "First, add a safe account using command" in result.output
    assert "ape safe add" in result.output


def test_list_no_txns(runner, cli, safe_account, chain):
    arguments = ("pending", "list", "--network", chain.provider.network_choice)
    result = runner.invoke(cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "There are no pending transactions" in result.output


def test_approve_transaction_not_found(runner, cli, safe_account, chain):
    tx_hash = "0x123"
    arguments = ("pending", "approve", tx_hash, "--network", chain.provider.network_choice)
    result = runner.invoke(
        cli,
        arguments,
        catch_exceptions=False,
    )
    assert result.exit_code != 0, result.output
    assert f"Pending transaction(s) '{tx_hash}' not found." in result.output


def test_approve(receiver, runner, cli, safe_account, chain):
    # First, fund the safe so the tx does not fail.
    receiver.transfer(safe_account, "1 ETH")
    tx_hash = "0x123"
    nonce = 1

    safe_account.client.transactions_by_nonce[nonce] = tx_hash
    safe_account.client.transactions[tx_hash] = ExecutedTxData(
        executionDate=datetime.now(),
        blockNumber=0,
        transactionHash=tx_hash,
        executor=receiver.address,
        isExecuted=False,
        isSuccessful=True,
        ethGasPrice=0,
        maxFeePerGas=1000,
        maxPriorityFeePerGas=1000,
        gasUsed=100,
        fee=10,
        origin="ape",
        dataDecoded=None,
        confirmationsRequired=0,
        safeTxHash=tx_hash,
        submissionDate=datetime.now(),
        modified=datetime.now(),
        nonce=nonce,
        refundReceiver=receiver.address,
        gasPrice=0,
        baseGas=0,
        safeTxGas=0,
        gasToken=receiver.address,
        operation=0,
        value=0,
        to=receiver.address,
        safe=safe_account.address,
    )

    arguments = ("pending", "approve", tx_hash, "--network", chain.provider.network_choice)
    result = runner.invoke(
        cli,
        arguments,
        catch_exceptions=False,
    )
    assert result.exit_code != 0, result.output



import time
from abc import ABC, abstractmethod
from collections.abc import Iterator
from functools import cached_property
from typing import TYPE_CHECKING, Optional, Union

import certifi
import requests
import urllib3
from ape.types import AddressType, HexBytes, MessageSignature
from eth_utils import keccak
from requests.adapters import HTTPAdapter

from ape_safe.client.types import (
    ExecutedTxData,
    SafeApiTxData,
    SafeDetails,
    SafeTx,
    SafeTxConfirmation,
    SafeTxID,
    UnexecutedTxData,
)
from ape_safe.exceptions import ClientResponseError

if TYPE_CHECKING:
    from ape.api import AccountAPI
    from requests import Response

DEFAULT_HEADERS = {
    "Accept": "application/json",
    "Content-Type": "application/json",
}


class BaseSafeClient(ABC):
    def __init__(self, base_url: str):
        self.base_url = base_url

    """Abstract methods"""

    @property
    @abstractmethod
    def safe_details(self) -> SafeDetails: ...

    @abstractmethod
    def get_next_nonce(self) -> int: ...

    @abstractmethod
    def _all_transactions(self) -> Iterator[SafeApiTxData]: ...

    @abstractmethod
    def get_safe_tx(self, safe_tx_hash: SafeTxID) -> SafeApiTxData: ...

    @abstractmethod
    def get_confirmations(self, safe_tx_hash: SafeTxID) -> Iterator[SafeTxConfirmation]: ...

    @abstractmethod
    def post_transaction(
        self, safe_tx: SafeTx, signatures: dict[AddressType, MessageSignature], **kwargs
    ): ...

    @abstractmethod
    def post_signatures(
        self,
        safe_tx_or_hash: Union[SafeTx, SafeTxID],
        signatures: dict[AddressType, MessageSignature],
    ): ...

    @abstractmethod
    def estimate_gas_cost(
        self, receiver: AddressType, value: int, data: bytes, operation: int = 0
    ) -> Optional[int]: ...

    """Shared methods"""

    def get_transactions(
        self,
        confirmed: Optional[bool] = None,
        starting_nonce: int = 0,
        ending_nonce: Optional[int] = None,
        filter_by_ids: Optional[set[SafeTxID]] = None,
        filter_by_missing_signers: Optional[set[AddressType]] = None,
    ) -> Iterator[SafeApiTxData]:
        """
        confirmed: Confirmed if True, not confirmed if False, both if None
        """
        next_nonce = self.get_next_nonce()

        # NOTE: We loop backwards.
        for txn in self._all_transactions():
            if ending_nonce is not None and txn.nonce > ending_nonce:
                # NOTE: Skip all largest nonces first
                continue

            elif txn.nonce < starting_nonce:
                break  # NOTE: order is largest nonce to smallest, so safe to break here

            is_confirmed = len(txn.confirmations) >= txn.confirmations_required

            if confirmed is not None:
                if not confirmed and isinstance(txn, ExecutedTxData):
                    break  # NOTE: Break at the first executed transaction

                elif confirmed and not is_confirmed:
                    continue  # NOTE: Skip not confirmed transactions

            # NOTE: use `type(txn) is ...` because ExecutedTxData is a subclass of UnexecutedTxData
            if txn.nonce < next_nonce and type(txn) is UnexecutedTxData:
                continue  # NOTE: Skip orphaned transactions

            if filter_by_ids and txn.safe_tx_hash not in filter_by_ids:
                continue  # NOTE: Skip transactions not in the filter

            if filter_by_missing_signers and filter_by_missing_signers.issubset(
                set(conf.owner for conf in txn.confirmations)
            ):
                # NOTE: Skip if all signers from `filter_by_missing_signers`
                #       are in `txn.confirmations`
                continue

            yield txn

    def create_delegate_message(self, delegate: AddressType) -> HexBytes:
        # NOTE: referencing https://github.com/safe-global/safe-eth-py/blob/
        # a0a5771622f143ee6301cfc381c5ed50832ff482/gnosis/safe/api/transaction_service_api.py#L34
        totp = int(time.time()) // 3600
        return HexBytes(keccak(text=(delegate + str(totp))))

    @abstractmethod
    def get_delegates(self) -> dict[AddressType, list[AddressType]]: ...

    @abstractmethod
    def add_delegate(self, delegate: AddressType, label: str, delegator: "AccountAPI"): ...

    @abstractmethod
    def remove_delegate(self, delegate: AddressType, delegator: "AccountAPI"): ...

    """Request methods"""

    @cached_property
    def session(self) -> requests.Session:
        session = requests.Session()
        adapter = HTTPAdapter(
            pool_connections=10,  # Doing all the connections to the same url
            pool_maxsize=100,  # Number of concurrent connections
            pool_block=False,
        )
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        return session

    def _get(self, url: str, params: Optional[dict] = None, **kwargs) -> "Response":
        return self._request("GET", url, params=params, **kwargs)

    def _post(self, url: str, json: Optional[dict] = None, **kwargs) -> "Response":
        return self._request("POST", url, json=json, **kwargs)

    def _delete(self, url: str, json: Optional[dict] = None, **kwargs) -> "Response":
        return self._request("DELETE", url, json=json, **kwargs)

    @cached_property
    def _http(self):
        return urllib3.PoolManager(ca_certs=certifi.where())

    def _request(self, method: str, url: str, json: Optional[dict] = None, **kwargs) -> "Response":
        api_version = kwargs.pop("api_version", "v1")

        # NOTE: paged requests include full url already
        if url.startswith(self.base_url):
            api_url = url

        else:
            api_url = f"{self.base_url}/{api_version}{url}"

        do_fail = not kwargs.pop("allow_failure", False)

        # Use `or 10` to handle when None is explicit.
        kwargs["timeout"] = kwargs.get("timeout") or 10

        # Add default headers
        headers = kwargs.get("headers", {})
        kwargs["headers"] = {**DEFAULT_HEADERS, **headers}
        response = self.session.request(method, api_url, json=json, **kwargs)

        if not response.ok and do_fail:
            raise ClientResponseError(api_url, response)

        return response
